"""diverse_seq: a tool for sampling diverse biological sequences"""

# need to import hdf5plugin here to make sure the plugin path can be
# found by h5py
import hdf5plugin  # noqa

__version__ = "2024.8.26a6"
